@@ -32,6 +32,9 @@
 
 #define MIME_ERROR g_quark_from_static_string("mime-error-quark")
 
+#define NOGLOBS "__NOGLOBS__"
+#define NOMAGIC "__NOMAGIC__"
+
 #ifndef PATH_SEPARATOR
 # ifdef _WIN32
 #  define PATH_SEPARATOR ";"
@@ -90,12 +93,15 @@ struct _Glob {
 	int weight;
 	char *pattern;
 	Type *type;
+	gboolean noglob;
+	gboolean case_sensitive;
 };
 
 struct _Magic {
 	int priority;
 	Type *type;
 	GList *matches;
+	gboolean nomagic;
 };
 
 struct _Match {
@@ -157,6 +163,7 @@ static GLogLevelFlags enabled_log_levels = G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITI
 
 /* Static prototypes */
 static Magic *magic_new(xmlNode *node, Type *type, GError **error);
+static Match *match_new(void);
 
 static TreeMagic *tree_magic_new(xmlNode *node, Type *type, GError **error);
 
@@ -301,6 +308,24 @@ static int get_weight(xmlNode *node)
        return get_int_attribute (node, "weight");
 }
 
+/* Return the value of a false/true attribute, which defaults to false.
+ * Returns 0 or 1.
+ */
+static gboolean get_boolean_attribute(xmlNode *node, const char* name)
+{
+	char *attr;
+	attr = my_xmlGetNsProp(node, name, NULL);
+	if (attr)
+	{
+	    if (strcmp (attr, "true") == 0) 
+	    {
+		return TRUE;
+	    }
+	    xmlFree(attr);
+	}
+	return FALSE;
+}
+
 /* Process a <root-XML> element by adding a rule to namespace_hash */
 static void add_namespace(Type *type, const char *namespaceURI,
 			  const char *localName, GError **error)
@@ -350,12 +375,17 @@ static void add_namespace(Type *type, const char *namespaceURI,
 static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 					 GError **error)
 {
+	gboolean copy_to_xml;
+
+	copy_to_xml = FALSE;
 	if (strcmp((char *)field->name, "glob") == 0)
 	{
 		gchar *pattern;	
 		gint weight;
+		gboolean case_sensitive;
 
 		weight = get_weight(field);
+		case_sensitive = get_boolean_attribute(field, "case-sensitive");
 
 		if (weight == -1)
 		{
@@ -373,9 +403,11 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 			glob->pattern = g_strdup (pattern);
 			glob->type = type;
 			glob->weight = weight;
+			glob->case_sensitive = case_sensitive;
 			list = g_list_append (list, glob);
 			g_hash_table_insert(globs_hash, g_strdup (glob->pattern), list);
 			xmlFree(pattern);
+			copy_to_xml = TRUE;
 		}
 		else
 		{
@@ -386,6 +418,21 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 				  "element"));
 		}
 	}
+	else if (strcmp((char *)field->name, "glob-deleteall") == 0)
+	{
+		Glob *glob;
+		GList *list = g_hash_table_lookup (globs_hash, NOGLOBS);
+
+		glob = g_new0 (Glob, 1);
+		glob->pattern = g_strdup (NOGLOBS);
+		glob->type = type;
+		glob->weight = 0;
+		glob->noglob = TRUE;
+		glob->case_sensitive = FALSE;
+		list = g_list_append (list, glob);
+		g_hash_table_insert(globs_hash, g_strdup (glob->pattern), list);
+		copy_to_xml = TRUE;
+	}
 	else if (strcmp((char *)field->name, "magic") == 0)
 	{
 		Magic *magic;
@@ -400,6 +447,22 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 		else
 			g_return_val_if_fail(magic == NULL, FALSE);
 	}
+	else if (strcmp((char *)field->name, "magic-deleteall") == 0)
+	{
+		Magic *magic;
+		Match *match;
+
+		magic = g_new0(Magic, 1);
+		magic->priority = 0;
+		magic->type = type;
+		magic->nomagic = TRUE;
+		match = match_new ();
+		match->data = g_strdup (NOMAGIC);
+		match->data_length = strlen (NOMAGIC);
+		magic->matches = g_list_prepend (NULL, match);
+
+		g_ptr_array_add(magic_array, magic);
+	}
 	else if (strcmp((char *)field->name, "treemagic") == 0)
 	{
 		TreeMagic *magic;
@@ -417,7 +480,7 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 	else if (strcmp((char *)field->name, "comment") == 0 ||
 		 strcmp((char *)field->name, "acronym") == 0 ||
 		 strcmp((char *)field->name, "expanded-acronym") == 0)
-		return FALSE;
+		copy_to_xml = TRUE;
 	else if (strcmp((char *)field->name, "alias") == 0 ||
 		 strcmp((char *)field->name, "sub-class-of") == 0)
 	{
@@ -445,18 +508,20 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 				nlist = g_slist_append (list, g_strdup(other_type));
 				if (list == NULL)
 					g_hash_table_insert(subclass_hash, 
-							    g_strdup(typename), nlist);    
+							    g_strdup(typename), nlist);
 			}
 			g_free(typename);
 			xmlFree(other_type);
-			
-			return FALSE;	/* Copy through */
+
+			copy_to_xml = TRUE; /* Copy through */
+		}
+		else
+		{
+			xmlFree(other_type);
+			g_set_error(error, MIME_ERROR, 0,
+				    _("Incorrect or missing 'type' attribute "
+				      "in <%s>"), field->name);
 		}
-		
-		xmlFree(other_type);
-		g_set_error(error, MIME_ERROR, 0,
-			_("Incorrect or missing 'type' attribute "
-			  "in <%s>"), field->name);
 	}
 	else if (strcmp((char *)field->name, "root-XML") == 0)
 	{
@@ -495,11 +560,13 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 
 			xmlFree (icon);
 
-			return FALSE;   /* Copy through */
+			copy_to_xml = TRUE; /* Copy through */
 		}
 	}
 
-	return !*error;
+	if (*error)
+		return FALSE;
+	return !copy_to_xml;
 }
 
 /* Checks to see if 'node' has the given value for xml:lang.
@@ -803,6 +870,7 @@ static void write_out_glob2(GList *globs, FILE *stream)
 	GList *list;
 	Glob *glob;
 	Type *type;
+	gboolean need_flags;
 
 	for (list = globs ; list; list = list->next) {
 		glob = (Glob *)list->data;
@@ -811,9 +879,22 @@ static void write_out_glob2(GList *globs, FILE *stream)
 			g_warning("* Glob patterns can't contain literal newlines "
 				  "(%s in type %s/%s)\n", glob->pattern,
 				  type->media, type->subtype);
-		else
+		else 
+		{
+			need_flags = FALSE;
+			if (glob->case_sensitive)
+				need_flags = TRUE;
+
+			if (need_flags) {
+				g_fprintf(stream, "%d:%s/%s:%s%s\n",
+						  glob->weight, type->media, type->subtype, glob->pattern,
+						  glob->case_sensitive ? ":cs" : "");
+			}
+
+			/* Always write the line without the flags, for older parsers */
 			g_fprintf(stream, "%d:%s/%s:%s\n",
-				  glob->weight, type->media, type->subtype, glob->pattern);
+					  glob->weight, type->media, type->subtype, glob->pattern);
+		}
 	}
 }
 
@@ -824,11 +905,14 @@ static void collect_glob2(gpointer key, gpointer value, gpointer data)
 	*listp = g_list_concat (*listp, g_list_copy ((GList *)value));
 }
 
-static int compare_by_weight (gpointer a, gpointer b)
+static int compare_glob_by_weight (gpointer a, gpointer b)
 {
 	Glob *ag = (Glob *)a;
 	Glob *bg = (Glob *)b;
 
+	if (ag->noglob || bg->noglob)
+		return bg->noglob - ag->noglob;
+
 	return bg->weight - ag->weight;
 }
 
@@ -887,6 +971,10 @@ static gint cmp_magic(gconstpointer a, gconstpointer b)
 	Magic *bb = *(Magic **) b;
 	int retval;
 
+	/* Sort nomagic items at start */
+	if (aa->nomagic || bb->nomagic)
+		return bb->nomagic - aa->nomagic;
+
 	if (aa->priority > bb->priority)
 		return -1;
 	else if (aa->priority < bb->priority)
@@ -1440,7 +1528,7 @@ static Magic *magic_new(xmlNode *node, Type *type, GError **error)
 	}
 	else
 	{
-		magic = g_new(Magic, 1);
+		magic = g_new0(Magic, 1);
 		magic->priority = prio;
 		magic->type = type;
 		magic->matches = build_matches(node, error);
@@ -2090,7 +2178,7 @@ write_card32 (FILE *cache, guint32 n)
 }
 
 #define MAJOR_VERSION 1
-#define MINOR_VERSION 1
+#define MINOR_VERSION 2
 
 static gboolean
 write_header (FILE *cache,   
@@ -2247,6 +2335,17 @@ get_type_value (gpointer  data,
   return result;
 }
 
+static guint32
+get_glob_weight_and_flags (Glob *glob)
+{
+  guint32 res;
+
+  res = glob->weight & 0xff;
+  if (glob->case_sensitive)
+    res |= 0x100;
+  return res;
+}
+
 static gchar **
 get_glob_list_value (gpointer  data, 
 		     gchar    *key)
@@ -2269,7 +2368,7 @@ get_glob_list_value (gpointer  data,
 
       result[i++] = g_strdup (glob->pattern);
       result[i++] = g_strdup_printf ("%s/%s", type->media, type->subtype);
-      result[i++] = g_strdup_printf ("%d", glob->weight);
+      result[i++] = g_strdup_printf ("%ud", get_glob_weight_and_flags (glob));
     }
   return result;
 }
@@ -2450,6 +2549,7 @@ struct _SuffixEntry
   gunichar character;
   gchar *mimetype;
   gint weight;
+  guint32 flags;
   GList *children;
   guint size;
   guint depth;
@@ -2458,7 +2558,8 @@ struct _SuffixEntry
 static GList *
 insert_suffix (gunichar *suffix, 
 	       gchar    *mimetype,
-	       gint      weight, 
+	       gint      weight,
+	       guint32   flags,
 	       GList    *suffixes)
 {
   GList *l;
@@ -2517,12 +2618,13 @@ insert_suffix (gunichar *suffix,
 	  s2->character = 0;
 	  s2->mimetype = mimetype;
 	  s2->weight = weight;
+	  s2->flags = flags;
 	  s2->children = NULL;
 	  s->children = g_list_insert_before (s->children, l2, s2);
 	}
-    } 
+    }
   else
-    s->children = insert_suffix (suffix + 1, mimetype, weight, s->children);
+    s->children = insert_suffix (suffix + 1, mimetype, weight, flags, s->children);
 
   return suffixes;
 }
@@ -2554,6 +2656,7 @@ build_suffixes (gpointer key,
   Glob *glob;
   Type *type;
   glong len;
+  guint32 flags;
   
   if (is_simple_glob (pattern))
     {
@@ -2572,7 +2675,10 @@ build_suffixes (gpointer key,
           type = glob->type;
           mimetype = g_strdup_printf ("%s/%s", type->media, type->subtype);
 
-          *suffixes = insert_suffix (suffix, mimetype, glob->weight, *suffixes);
+	  flags = 0;
+	  if (glob->case_sensitive)
+	    flags |= 0x100;
+          *suffixes = insert_suffix (suffix, mimetype, glob->weight, flags, *suffixes);
         }
 
       g_free (suffix);
@@ -2640,7 +2746,7 @@ write_suffix_entries (FILE        *cache,
       if (!write_card32 (cache, offset))
         return FALSE;
 
-      if (!write_card32 (cache, entry->weight))
+      if (!write_card32 (cache, (entry->weight & 0xff) | entry->flags))
         return FALSE;
     }
   else
@@ -3386,6 +3492,12 @@ int main(int argc, char **argv)
 
 	package_dir = g_strconcat(mime_dir, "/packages", NULL);
 
+	if (access(mime_dir, F_OK))
+	{
+		g_warning(_("Directory '%s' does not exist!\n"), package_dir);
+		return EXIT_FAILURE;
+	}
+
 	if (access(mime_dir, W_OK))
 	{
 		g_warning(_("%s: I don't have write permission on %s.\n"
@@ -3432,7 +3544,7 @@ int main(int argc, char **argv)
 		GList *glob_list = NULL;
 
 		g_hash_table_foreach(globs_hash, collect_glob2, &glob_list);
-		glob_list = g_list_sort(glob_list, (GCompareFunc)compare_by_weight);
+		glob_list = g_list_sort(glob_list, (GCompareFunc)compare_glob_by_weight);
 		globs_path = g_strconcat(mime_dir, "/globs.new", NULL);
 		globs = open_or_die(globs_path);
 		g_fprintf(globs,
